import numpy as np


def para_ortho_equilibrium(T):
    # Calculate the equilibrium concentrations of Para- & Ortho- Hydrogen as a
    # function of temperature (T)
    xp = 0.1*(np.exp(-175./T)+0.1)**(-1) - 7.06e-9 * \
        T**3 + 3.42e-6*T**2 - 6.2e-5*T - 0.00227
    xo = 1 - xp
    return xp, xo


def first_order_kinetics(reactant):
    # calculate first order conversion kinetics for ortho > para conversion
    # https://onlinelibrary.wiley.com/doi/abs/10.1002/ceat.201800345
    # extract reactant properties.
    T = reactant.temperature
    xo = reactant.get_xo()
    xo_equil = reactant.get_xo_equil()
    cH2 = reactant.molar_density

    # universal gas constant
    # TODO - import from CP
    R = 8.31446261815324  # J/(mol.K)

    # reactor kinetic constants (from Donaubauer et al.)
    Ea = -336.45                # J/mol
    a = 2.2e-3                 # m3s/mol
    b = -35.11e-3              # s

    # rate constant
    k = np.exp(-Ea/(R*T))/(a*cH2 + b) * 1/(1-xo_equil)  # 1/s

    # first order conversion rate. Note conversion rate defined in terms of
    # production of para-hydrogen as positive.
    conversion_rate = k * cH2 * (xo - xo_equil)  # m3 / (mol.s)

    return conversion_rate



import numpy as np
from hydrogen_pfhx import heat_transfer_models, pressure_models


class PlateFinHex(object):
    def __init__(self, reactor_configuration):
        # keeping constant for now
        self.hot_fraction = 0.5

        # input
        self.length = reactor_configuration['length']
        self.width = reactor_configuration['width']
        self.height = reactor_configuration['height']
        self.fin_thickness = reactor_configuration['fin_thickness']
        self.fin_pitch = reactor_configuration['fin_pitch']  # (in width direction)
        self.fin_height = reactor_configuration['fin_height']    # (in height direction)
        self.seration_length = reactor_configuration['seration_length']

    def characteristic_length(self):
        cl = self.hydraulic_diameter
        return cl

    def calculate_fin_efficiency(self, h, k):
        I = self.fin_height/2 - self.fin_thickness
        m = (2*h/(k*self.fin_thickness) * (1 + self.fin_thickness/I))**0.5
        fe = np.tanh(m*I)/(m*I)
        return fe

    def fin_spacing(self):
        s = self.fin_pitch - self.fin_thickness
        return s

    def layer_height(self):
        lh = self.fin_height + self.fin_thickness
        return lh

    def number_layers(self):
        # ensuring even number of layers (for hot/cold)
        nl = np.floor(self.height / self.layer_height()/2)*2
        return nl

    def hydraulic_perimeter(self):
        ph = 2*(self.fin_height + self.fin_spacing())
        return ph

    def hydraulic_diameter(self):
        dh = 4*self.single_channel_area() / self.hydraulic_perimeter()
        return dh

    def total_channels(self):
        channels_per_layer = np.floor(self.width / self.fin_pitch)
        tc = channels_per_layer * self.number_layers()
        return tc

    def hot_channels(self):
        hc = self.hot_fraction * self.total_channels()
        return hc

    def cold_channels(self):
        cc = (1-self.hot_fraction)*self.total_channels()
        return cc

    def single_channel_area(self):
        csa = self.fin_height * self.fin_spacing()
        return csa

    def total_channel_area(self):
        tca = self.single_channel_area() * self.total_channels()
        return tca

    def total_cold_side_area(self):
        tcca = self.single_channel_area() * self.cold_channels()
        return tcca

    def total_hot_side_area(self):
        thca = self.single_channel_area() * self.hot_channels()
        return thca

    def calculate_heat_transfer_duty(self, reactant_temperature, coolant_temperature):
        wall_temperature = (coolant_temperature + reactant_temperature) / 2
        (U, terms) = self.calculate_overall_heat_transfer_coefficient(
            wall_temperature)
        Q = U*self.hydraulic_perimeter()*(reactant_temperature - coolant_temperature) * \
            self.hot_channels()  # W/m
        return Q, terms

    def hot_side_pressure_drop(self, reactant, catalyst):
        voidage = catalyst.void_fraction
        vs = voidage * reactant.velocity
        Dp = catalyst.particle_diameter
        rho = reactant.mass_density
        delta_p = -(150*reactant.viscosity * (1-voidage)**2 * vs / (Dp**2 *
                    voidage**3) + 1.75 * rho * (1 - voidage) * vs**2 / (Dp * voidage ** 3))
        return delta_p

    def cold_side_pressure_drop(self, coolant):
        # non-negative due to direction
        delta_p = 2 * self.f_cold * coolant.mass_density * \
            coolant.velocity**2 / (self.hydraulic_diameter())
        return delta_p

    def calculate_overall_heat_transfer_coefficient(self, wall_temperatures):
        solid_thermal_conductivity = np.mean(
            heat_transfer_models.aluminium_thermal_conductivity(wall_temperatures))
        term_1 = self.hydraulic_diameter()/(8*self.k_hot)
        term_2 = 1/(self.h_hot)
        term_3 = self.fin_thickness / solid_thermal_conductivity
        term_4 = 1/(self.h_cold)
        terms = [term_1, term_2, term_3, term_4]
        u = 1/np.sum(terms)
        return u, terms

    # friction factor (f) & heat transfer coefficient (h) for hot side.
    def hot_side_transport_behaviour(self, reactant, catalyst, wall_temperature):
        # initial transport properties
        self.k_hot = self.effective_radial_conductivity(
            reactant, catalyst, self.hydraulic_diameter())
        Pr_eff = reactant.viscosity * reactant.specific_heat_capacity / \
            reactant.thermal_conductivity  # self.k_hot

        # wall properties
        original_temperature = reactant.temperature
        reactant.update_conditions(wall_temperature, reactant.pressure)
        k_wall = self.effective_radial_conductivity(
            reactant, catalyst, self.hydraulic_diameter())
        Pr_w = reactant.viscosity * reactant.specific_heat_capacity / k_wall
        reactant.update_conditions(original_temperature, reactant.pressure)

        # Reynolds number
        reactant.calculate_reynolds_number(catalyst.particle_diameter)
        Re = reactant.reynolds_number
        aspect_ratio = self.hydraulic_diameter() / self.length

        # friction factor
        self.f_hot = pressure_models.ergun_equation(catalyst.void_fraction, Re)

        # Nusselt number correlations
        if Re < 0:
            print('Cannot have negative Reynolds')
        elif Re <= 2000:
            Nu_h = heat_transfer_models.laminar_nusselt_correlation(
                aspect_ratio)
        elif (Re > 2000) & (Re < 8000):
            f_darcy = pressure_models.ergun_equation(
                catalyst.void_fraction, Re)
            Nu_lam = heat_transfer_models.laminar_nusselt_correlation(
                aspect_ratio)

            Nu_turb = heat_transfer_models.gnielinski_equation(
                f_darcy, Re, Pr_eff, Pr_w, aspect_ratio)
            phi = 4/3 - Re/6000
            Nu_h = phi*Nu_turb + (1-phi)*Nu_lam
        else:
            f_darcy = pressure_models.ergun_equation(
                catalyst.void_fraction, Re)
            Nu_h = heat_transfer_models.gnielinski_equation(
                f_darcy, Re, Pr_eff, Pr_w, aspect_ratio)

        h_h = Nu_h*self.k_hot/self.hydraulic_diameter()

        # correct due to fin efficiency
        fe_hot = self.calculate_fin_efficiency(
            h_h, reactant.thermal_conductivity)
        self.h_hot = fe_hot*h_h


    def cold_side_transport_behaviour(self, coolant):
        # friction factor (f) & heat transfer coefficient (h) for cold side.
        alpha = self.fin_spacing() / self.fin_height
        delta = self.fin_thickness / self.seration_length
        gamma = self.fin_thickness / self.fin_spacing()
        (f, j) = heat_transfer_models.manglik_bergles_heat_transfer_model(
            coolant.reynolds_number, alpha, delta, gamma)
        self.f_cold = f
        Nu_c = j * coolant.reynolds_number * coolant.prandtl_number**(1/3)
        h_c = Nu_c*coolant.thermal_conductivity/self.hydraulic_diameter()

        # correct due to fin efficiency
        fe_cold = self.calculate_fin_efficiency(
            h_c, coolant.thermal_conductivity)
        h_c = h_c * fe_cold
        self.h_cold = h_c

    def effective_radial_conductivity(self, reactant, catalyst, d_t):
        # props
        k_f = reactant.thermal_conductivity
        k_s = catalyst.calculate_thermal_conductivity(reactant.temperature)
        epsilon = catalyst.void_fraction
        dp = catalyst.particle_diameter

        zeta = 0.9
        alpha_rs = 2e-5 * (zeta/(2-zeta))*(reactant.temperature/100)**3
        B = 1.25 * ((1 - epsilon) / epsilon)**(10/9)

        k_r = k_f/k_s
        dim_alpha = alpha_rs * dp / k_f

        omega = (1+(dim_alpha - 1)*k_r)/(1+(dim_alpha - B)*k_r) * np.log((1 + alpha_rs*dp/k_s) /
                                                                         (B*k_r)) - (B-1) / (1 + (dim_alpha - B)*k_r) + (B+1) / (2*B) * (dim_alpha - B)
        k_0 = k_f * ((1 - (1-epsilon))**0.5 * (1 + epsilon * dim_alpha) +
                     2*omega * (1-epsilon)**0.5/(1+(dim_alpha - B)*k_r))

        # dynamic thermal conductivity
        Pe = (1 + 46*(dp/d_t)**2)/0.14  # radial Peclet number
        D_er = reactant.velocity * dp / (epsilon * Pe)
        k_t = epsilon * reactant.mass_density * reactant.specific_heat_capacity * D_er

        k_eff = k_0 + k_t
        return k_eff


"""
functions to model boundary value problem for PFHX
"""

import numpy as np
from hydrogen_pfhx import ortho_para_dynamics

# Main 1D BVP model of h2 conversion reactor.
def bvp_function(z, process_properties, additional_parameters):
    # extract the "additional parameters"
    # Note - must be organised in this way due to bvp_solver setup.
    reactor = additional_parameters[0]
    catalyst = additional_parameters[1]
    reactant = additional_parameters[2]
    coolant = additional_parameters[3]
    boundary_properties = additional_parameters[4]

    # (reactor, catalyst, reactant, coolant, boundary_properties)
    delta_process_properties = np.zeros(process_properties.shape)
    nodes = len(z)

    term_matrix = np.zeros((nodes, 4))
    for zi in np.arange(nodes):
        current_process_properties = process_properties[:, zi]
        # update the conditions (T & P) of the fluid streams. Also update reacting
        # stream composition.
        [reactant, coolant] = update_parameters(
            current_process_properties, reactant, coolant, boundary_properties)

        # calculate wall temperature
        wall_temperature = (coolant.temperature + reactant.temperature)/2

        # Reactant side calcs
        reactant.calculate_velocity(
            reactor.total_hot_side_area()*catalyst.void_fraction)
        reactant.calculate_reynolds_number(
            catalyst.particle_diameter/catalyst.solid_fraction)
        reactor.hot_side_transport_behaviour(
            reactant, catalyst, wall_temperature)

        # coolant side calcs
        coolant.calculate_velocity(reactor.total_cold_side_area())
        coolant.calculate_reynolds_number(reactor.hydraulic_diameter())
        reactor.cold_side_transport_behaviour(coolant)

        # reactor conversion
        r_dot = ortho_para_dynamics.first_order_kinetics(
            reactant)              # mol/m3/s
        dNpdz = r_dot * reactor.total_hot_side_area()*catalyst.solid_fraction   # mol/m/s
        dxpdz = dNpdz * reactant.molecular_mass / reactant.mass_flow_rate

        # heat transfer calculations
        (Q_transfer, _) = reactor.calculate_heat_transfer_duty(
            reactant.temperature, coolant.temperature)
        hoc = reactant.get_heat_of_conversion()
        Q_hot = -Q_transfer - hoc * dNpdz

        # negative due to direction
        dTcdz = -Q_transfer/(coolant.mass_flow_rate *
                             coolant.specific_heat_capacity)
        # Q_hot is negative as heat transferred out of reactant
        dTrdz = 1/(reactant.mass_flow_rate *
                   reactant.specific_heat_capacity)*Q_hot

        # Basic assumption that conversion in coolant is extremely slow - needs to be
        # updated.
        dPrdz = reactor.hot_side_pressure_drop(reactant, catalyst)
        dPcdz = reactor.cold_side_pressure_drop(coolant)
        current_deltas = np.vstack(
            ([dxpdz[0], dPrdz[0]/1e3, dTrdz[0], dPcdz/1e3, dTcdz[0]]))
        delta_process_properties[:, zi] = current_deltas.flatten()

    return delta_process_properties

# define boundary conditions
def counter_current_boundary_condition(inlet_properties, outlet_properties, boundary_parameters):
    # corresponds to reactant inlet (x_para, P & T)
    inlet_residual = inlet_properties[0:3] - boundary_parameters[0:3]
    # corresponds to coolant inlet (P & T)
    outlet_residual = outlet_properties[3:5] - boundary_parameters[3:5]
    bc_res = np.hstack([inlet_residual, outlet_residual])
    return bc_res

# method for defining an initial guess for the solution.
def initialise_solution(reactant, coolant, reactor, catalyst, boundary_properties, simulation_configuration):
    nodes = simulation_configuration['nodes']
    x_linear = np.linspace(0, reactor.length, nodes)
    delta_T_guess = simulation_configuration['delta_t']
    T_reactant_outlet = boundary_properties[4] + delta_T_guess
    T_coolant_outlet = boundary_properties[2] - delta_T_guess
    sol_guess = initial_guess(x_linear, boundary_properties,
                              reactor.length, T_reactant_outlet, T_coolant_outlet)
    reactant_tc = np.zeros((nodes,))
    reactant_cp = np.zeros((nodes,))
    coolant_tc = np.zeros((nodes,))
    coolant_cp = np.zeros((nodes,))
    reactant_h = np.zeros((nodes,))
    reactant_T = np.zeros((nodes,))
    coolant_h = np.zeros((nodes,))
    coolant_T = np.zeros((nodes,))

    for xi in np.arange(nodes):
        (reactant, coolant) = update_parameters(
            sol_guess[:, xi], reactant, coolant, boundary_properties)
        reactant_tc[xi] = reactant.thermal_conductivity
        reactant_cp[xi] = reactant.specific_heat_capacity
        reactant_h[xi] = reactant.enthalpy
        reactant_T[xi] = reactant.temperature
        coolant_tc[xi] = coolant.thermal_conductivity
        coolant_cp[xi] = coolant.specific_heat_capacity
        coolant_h[xi] = coolant.enthalpy
        coolant_T[xi] = coolant.temperature

    reactant_q = reactant_h * reactant.mass_flow_rate
    coolant_q = coolant_h * coolant.mass_flow_rate
    reactant_duty = reactant_q[-1] - reactant_q[0]
    coolant_duty = coolant_q[-1] - coolant_q[0]
    if abs(reactant_duty) > abs(coolant_duty):
        outlet_reactant_h = reactant_h[0] + \
            coolant_duty/reactant.mass_flow_rate
        T_reactant_outlet = np.interp(
            outlet_reactant_h, reactant_h, reactant_T)
        T_coolant_outlet = boundary_properties[2] - delta_T_guess
    elif abs(reactant_duty) < abs(coolant_duty):
        outlet_coolant_h = coolant_h[-1] - reactant_duty/coolant.mass_flow_rate
        T_coolant_outlet = np.max(
            (boundary_properties[4], np.interp(outlet_coolant_h, coolant_h, coolant_T)))
        T_reactant_outlet = boundary_properties[4] + delta_T_guess

    sol_init = initial_guess(x_linear, boundary_properties,
                             reactor.length, T_reactant_outlet, T_coolant_outlet)

    return x_linear, sol_init


def initial_guess(x, bc, length, T_reactant_outlet, T_coolant_outlet):
    nodes = len(x)
    T_reactant_inlet = bc[2]
    T_coolant_inlet = bc[4]
    T_reactant_gradient = (T_reactant_outlet - T_reactant_inlet) / length
    T_reactant_profile = T_reactant_inlet + T_reactant_gradient*x
    (final_xp, final_xo) = ortho_para_dynamics.para_ortho_equilibrium(T_reactant_outlet)
    # TODO - abstract the dxp guess into config.
    xp_adj = final_xp - 0.05
    xp_profile = bc[0] + (xp_adj - bc[0])*x/length
    T_coolant_gradient = (T_coolant_inlet - T_coolant_outlet) / length
    T_coolant_profile = T_coolant_outlet + T_coolant_gradient*x
    P_coolant = bc[3]*np.ones(nodes,)
    sol = np.concatenate([np.reshape(xp_profile, (1, nodes)),
                          np.reshape(bc[1]*(1-0.003*x / length), (1, nodes)),
                          np.reshape(T_reactant_profile, (1, nodes)),
                          np.reshape(P_coolant, (1, nodes)),
                          np.reshape(T_coolant_profile, (1, nodes))],
                         axis=0)
    return sol


def update_parameters(process_properties, reactant, coolant, boundary_properties):
    # impose a lower limit (non-zero) on all values (else functions may fail)
    nz_limit = 1e-6
    # extract the process properties
    # upper limit of 1 placed on composition.
    xp_reactant = np.max((0.0, np.min((process_properties[0], 1.0))))
    P_reactant = np.max((nz_limit, process_properties[1]))*1e3
    T_reactant = np.max((nz_limit, process_properties[2]))

    if len(process_properties) > 3:
        P_coolant = np.max((nz_limit, process_properties[3]))*1e3
        T_coolant = np.max((boundary_properties[4], process_properties[4]))
        coolant.update_conditions(T_coolant, P_coolant)
        coolant.set_properties()
    else:
        T_coolant = coolant.temperature

    # reactant can't be less than coolant
    T_reactant = np.max((T_coolant, T_reactant))

    # This will update the properties of the fluid stream (e.g. density,
    # viscosity)
    reactant.update_conditions(T_reactant, P_reactant)
    reactant.update_composition(xp_reactant)
    reactant.set_properties()

    return reactant, coolant



import numpy as np
from hydrogen_pfhx import pressure_models


def aluminium_thermal_conductivity(temperature):
    # Aluminium 3003 thermal conductivity
    # Woodcraft, Adam L. "Predicting the thermal conductivity of aluminium alloys in the cryogenic to room temperature range." Cryogenics 45.6 (2005): 421-431.
    rho_rt_pure = 2.43e-8  # ohm-m
    L0 = 2.45e-8  # W ohm / K^2
    RRR_star = 2.65  # relative RRR for 3003
    rho_0 = rho_rt_pure / RRR_star
    beta = rho_0 / L0

    P = np.array([4.716e-8, 2.446, 623.6, -0.16, 130.9, 2.5, 0.8168])

    W0 = beta / temperature
    Wi_a = P[0]*temperature**P[1]
    Wi_b = (1 + P[0]*P[2] * temperature**(P[1] + P[3])
            * np.exp(-((P[4]/temperature)**P[5])))**(-1)
    Wi_c = -5e-4 * np.log(temperature/330) * np.exp(-(np.log(temperature/380)/0.6)**2) - \
        1.3e-3 * np.log(temperature/110) * \
        np.exp(-(np.log(temperature/94)/0.5)**2)
    Wi = Wi_a * Wi_b + Wi_c
    Wi0 = P[6] * Wi*W0/(Wi + W0)
    k = 1 / (W0 + Wi + Wi0)
    return k


def stainless_steel_thermal_conductivity(T):
    # https://trc.nist.gov/cryogenics/materials/304Stainless/304Stainless_rev.htm
    a = -1.4087
    b = 1.3982
    c = 0.2543
    d = -0.6260
    e = 0.2334
    f = 0.4256
    g = -0.4658
    h = 0.1650
    i = -0.0199
    log10_y = a+b*(np.log10(T)) + c*(np.log10(T)) ** 2 + d*(np.log10(T)) ** 3 + e*(np.log10(T)) ** 4 + \
        f*(np.log10(T)) ** 5 + g*(np.log10(T)) ** 6 + \
        h*(np.log10(T)) ** 7 + i*(np.log10(T)) ** 8
    y = 10**(log10_y)
    return y


def gnielinski_equation(f_darcy, Re, Pr, Pr_w, aspect_ratio):
    K = (Pr / Pr_w)**0.11
    if (Pr < 0.5) | (Pr > 2000):
        print('Current Prandtl number (%.2e) is outside correlation limits', Pr)

    if (Re < 2000) | (Re > 5e6):  # edited to 2000 to allow use in transition zone (with interpolation)
        print('Current Reynolds number (%.2e) is outside correlation limits', Re)

    numerator = (f_darcy/8)*Re*Pr
    denominator = 1 + 12.7*(f_darcy/8)**0.5 * (Pr**(2/3)-1)
    correction_term = 1 + (aspect_ratio)**(2/3)

    Nu = numerator / denominator * correction_term * K
    return Nu


def laminar_nusselt_correlation(aspect_ratio):
    # Fundamentals of Heat Exchanger Design
    # Shah, Ramesh K; Sekulic, Dusan P
    # Table 7.4 for Rectangular geometry, Nu_T correlation
    Nu = 7.451*(1-2.61*aspect_ratio + 4.97 * aspect_ratio**2 - 5.119 *
                aspect_ratio**3 + 2.702*aspect_ratio**4 - 0.548*aspect_ratio**5)
    return Nu

# https://www.sciencedirect.com/science/article/pii/S0017931013003207#e0035


def gnielinski_laminar(Re, Pr, d, L):
    Nu_mT1 = 3.66
    Nu_mT2 = 1.615 * (Re * Pr * (d/L))**(1/3)
    Nu_mT3 = (2 / (1+22*Pr))**(1/6) * (Re * Pr * d/L)**0.5
    Nu_mT = (Nu_mT1**3 + 0.7**3 + (Nu_mT2 - 0.7)**3 + Nu_mT3**3)**(1/3)
    return Nu_mT


def full_gnielinski(Re, Pr, d, L, eta, Pr_w):
    # Nusselt number correlations
    aspect_ratio = d/L
    if Re < 0:
        print('Cannot have negative Reynolds')
    elif Re <= 2300:
        Nu_h = gnielinski_laminar(Re, Pr, d, L)
    elif (Re > 2300) & (Re < 4000):
        f_darcy = pressure_models.ergun_equation(eta, Re)
        Nu_lam = gnielinski_laminar(Re, Pr, d, L)

        Nu_turb = gnielinski_equation(f_darcy, Re, Pr, Pr_w, aspect_ratio)
        phi = (Re-2300) / (4000-2300)
        Nu_h = phi*Nu_turb + (1-phi)*Nu_lam
    else:
        f_darcy = pressure_models.ergun_equation(eta, Re)
        Nu_h = gnielinski_equation(f_darcy, Re, Pr, Pr_w, aspect_ratio)

    return Nu_h


def manglik_bergles_heat_transfer_model(Re, alpha, delta, gamma):
    # https://www.sciencedirect.com/science/article/pii/089417779400096Q
    f = 9.6243 * Re**-0.7422 * alpha**-0.1856 * delta**0.3053 * gamma**0.2659 * \
        (1 + 7.669e-8 * Re**4.429 * alpha**0.92 * delta**3.767 * gamma**0.236)**0.1
    j = 0.6522 * Re**-0.5403 * alpha**-0.1541 * delta**0.1499 * gamma**-0.0678 * \
        (1 + 5.269e-5 * Re**1.340 * alpha**0.504 * delta**0.456 * gamma**-1.055)**0.1
    return f, j


def packed_bed_heat_transfer(Dp, Dt, Re_p, Pr):
    # Peters PE, Schiffino RS, Harriott P. Heat transfer in packed tube reactors. Ind Eng Chem Res 1988;27:226e33. https://doi.org/10.1021/ie00074a003
    Nu = 3.8*(Dp/Dt)**0.39*Re_p ^ 0.5*Pr**(1/3)
    return Nu


def zehner_bauer_schlunder_model(k_g, epsilon, **kwargs):
    # Zehner - Bauer - Schlunder model for heat transfer in packed bed.
    omega = 7.26e-3
    void = 1 - epsilon

    k_ge = (1 - epsilon ^ 0.5)*k_g/void

    if 'k_s' in kwargs:
        k_s = kwargs['k_s']  # solid phase thermal conductivity
        B = 1.25*(epsilon/void) ^ (10/9)
        kappa = k_s/k_g
        ombok = 1-B/kappa  # One minus beta over kappa
        gamma = 2/ombok * ((kappa-1)/ombok ^ 2 * B/kappa *
                           np.log(kappa/B) - (B-1)/ombok - 0.5 * (B+1))

        k_se = epsilon ^ 0.5 * (omega*kappa + (1-omega)*gamma) * k_g / epsilon
    return k_ge, k_se



import numpy as np
import CoolProp.CoolProp as CP
from hydrogen_pfhx import ortho_para_dynamics, hydrogen_thermodynamics


class FluidStream(object):
    # Defines a fluid stream which is an inlet for a conversion HEX.

    # Construct fluid stream
    #TODO - define mass flow rate elsewhere.
    def __init__(self, fluid, mass_flow_rate):
        # define input attributes
        self.fluid = fluid
        self.mass_flow_rate = mass_flow_rate

        # setup abstract state
        self.abstract_state = CP.AbstractState('HEOS', fluid)

        # Environmental variables
        self.temperature = None
        self.pressure = None
        self.phase_fraction = None

        # Derived properties
        self.phase_index = None             # see below
        self.molecular_mass = None
        self.molar_density = None                 # molar density
        self.mass_density = None
        self.viscosity = None               # dynamic
        self.specific_heat_capacity = None  # constant pressure
        self.thermal_conductivity = None
        self.prandtl_number = None
        self.enthalpy = None
        self.entropy = None
        self.saturation_pressure = None
        self.latent_heat_vaporisation = None
        self.molar_flow_rate = None
        self.liquid_fraction = None
        self.vapor_fraction = None

        self.velocity = None
        self.reynolds_number = None
        self.critical_temperature = None
        self.critical_pressure = None

    def update_conditions(self, temperature, pressure):
        self.temperature = temperature
        self.pressure = pressure
        self.abstract_state.update(CP.PT_INPUTS, pressure, temperature)
        # self.abstract_state.specify_phase(CP.iphase_supercritical_liquid)

    def specify_phase(self, phase):
        self.abstract_state.specify_phase(phase)

    def set_mole_fractions(self, z):
        self.mole_fractions = z
        self.abstract_state.set_mole_fractions(z)

    def set_properties(self):
        # Set all key properties for fluid.
        self.molecular_mass = self._calculate_molecular_mass()
        self.molar_density = self._calculate_molar_density()
        self.mass_density = self.molar_density * self.molecular_mass
        self.viscosity = self._calculate_viscosity()
        self.specific_heat_capacity = self._calculate_specific_heat_capacity()
        self.thermal_conductivity = self._calculate_thermal_conductivity()
        self.prandtl_number = self.viscosity * \
            self.specific_heat_capacity/self.thermal_conductivity
        self.enthalpy = self._calculate_enthalpy()
        self.entropy = self._calculate_entropy()
        self.molar_flow_rate = self.mass_flow_rate / self.molecular_mass
        self.critical_temperature = self._calculate_critical_temperature()
        self.critical_pressure = self._calculate_critical_pressure()

    def calculate_velocity(self, cross_sectional_area):
        self.velocity = self.mass_flow_rate / \
            (self.mass_density * cross_sectional_area)

    # Reynolds number

    def calculate_reynolds_number(self, characteristic_length):
        self.reynolds_number = self.mass_density * \
            self.velocity * characteristic_length / self.viscosity

    # base methods for evaluating derived properties (can be modified
    # by subclasses, e.g. Hydrogen with mixed para/ortho)

    def _calculate_molar_density(self):
        # if two-phase - then calculate contribution of each phase
        if self.phase_index == 6:
            # get liquid property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            liquid_density = self.abstract_state.rhomolar()

            # get vapor property
            self.abstract_state.specify_phase(CP.iphase_vapour)
            vapor_density = self.abstract_state.rhomolar()

            # calculate combined property
            rho = self.liquid_fraction * liquid_density + self.vapor_fraction * vapor_density

            # return to two phase state
            self.abstract_state.specify_phase(CP.iphase_twophase)
        else:
            rho = self.abstract_state.rhomolar()
            
        return rho

    def _calculate_viscosity(self):

        # if two-phase - then calculate contribution of each phase
        if self.phase_index == 6:
            # get liquid property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            liquid_viscosity = self.abstract_state.viscosity()

            # get vapor property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            vapor_viscosity = self.abstract_state.viscosity()

            # calculate combined property
            mu = self.liquid_fraction * liquid_viscosity + \
                self.vapor_fraction * vapor_viscosity

            # return to two phase state
            self.abstract_state.specify_phase(CP.iphase_twophase)
        else:
            mu = self.abstract_state.viscosity()
            
        return mu

    def _calculate_specific_heat_capacity(self):
        if self.phase_index == 6:
            # get liquid property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            liquid_cp = self.abstract_state.cpmass()

            # get vapor property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            vapor_cp = self.abstract_state.cpmass()

            # calculate combined property
            cp = self.liquid_fraction * liquid_cp + self.vapor_fraction * vapor_cp

            # return to two phase state
            self.abstract_state.specify_phase(CP.iphase_twophase)
        else:
            cp = self.abstract_state.cpmass()

        return cp

    def _calculate_thermal_conductivity(self):
        if self.phase_index == 6:
            # get liquid property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            liquid_conductivity = self.abstract_state.conductivity()

            # get vapor property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            vapor_conductivity = self.abstract_state.conductivity()

            # calculate combined property
            k = self.liquid_fraction * liquid_conductivity + \
                self.vapor_fraction * vapor_conductivity

            # return to two phase state
            self.abstract_state.specify_phase(CP.iphase_twophase)
        else:
            k = self.abstract_state.conductivity()

        return k

    def _calculate_enthalpy(self):
        if self.phase_index == 6:
            # get liquid property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            liquid_enthalpy = self.abstract_state.enthalpy()

            # get vapor property
            self.abstract_state.specify_phase(CP.iphase_liquid)
            vapor_enthalpy = self.abstract_state.enthalpy()

            # calculate combined property
            h = self.liquid_fraction * liquid_enthalpy + \
                self.vapor_fraction * vapor_enthalpy

            # return to two phase state
            self.abstract_state.specify_phase(CP.iphase_twophase)
        else:
            h = self.abstract_state.hmass()

        return h

    def _calculate_entropy(self):
        s = self.abstract_state.smass
        return s

    def _calculate_latent_heat_vaporisation(self):
        # adjust to saturated liquid
        self.abstract_state.update(CP.QT_INPUTS, 1, self.temperature)
        h_v = self.abstract_state.hmass()

        # adjust to saturated gas
        self.abstract_state.update(CP.QT_INPUTS, 0, self.temperature)
        h_l = self.abstract_state.hmass()

        # calculate latent heat of vaporisation
        lhv = h_v - h_l

        # return to initial conditions
        self.abstract_state.update(
            CP.PT_INPUTS, self.pressure, self.temperature)
        return lhv

    # single parameter property
    def _calculate_saturation_pressure(self):
        # adjust to saturated gas
        self.abstract_state.update(CP.QT_INPUTS, 0, self.temperature)
        p_sat = self.abstract_state.p()

        # return to initial conditions
        self.abstract_state.update(
            CP.PT_INPUTS, self.pressure, self.temperature)
        return p_sat

    # trivial properties
    def _calculate_molecular_mass(self):
        m = self.abstract_state.molar_mass()
        return m

    def _calculate_critical_temperature(self):
        Tc = self.abstract_state.T_critical()
        return Tc

    def _calculate_critical_pressure(self):
        Pc = self.abstract_state.p_critical()
        return Pc


class Hydrogen(FluidStream):
    # Hydrogen stream inlet into HEX
    def __init__(self, mass_flow_rate):
        FluidStream.__init__(self, 'Hydrogen', mass_flow_rate)
        self.ortho = hydrogen_thermodynamics.OrthoHydrogen()
        self.para = hydrogen_thermodynamics.ParaHydrogen()
        self.normal = hydrogen_thermodynamics.NormalHydrogen()

    def set_properties(self):
        # Only set the densities once per update in conditions. Need
        # to set before other properties so they can be used, but also
        # need to determine rho guess first.
        rho_guess = self.abstract_state.rhomolar()
        self.ortho_density = self.ortho.calculate_density(
            self.pressure, self.temperature, rho_guess)
        self.para_density = self.para.calculate_density(
            self.pressure, self.temperature, rho_guess)
        self.normal_density = self.normal.calculate_density(
            self.pressure, self.temperature, rho_guess)

        FluidStream.set_properties(self)

    def update_composition(self, xp):
        self.xp = xp

    def get_xp(self):
        xp = self.xp
        return xp

    def get_xo(self):
        xo = 1.0 - self.xp
        return xo

    def get_xp_equil(self):
        (xpe, _) = ortho_para_dynamics.para_ortho_equilibrium(self.temperature)
        return xpe

    def get_xo_equil(self):
        xpo = 1.0 - self.get_xp_equil()
        return xpo

    def get_heat_of_conversion(self):
        ortho_enthalpy = self.ortho.enthalpy(
            self.ortho_density, self.temperature)
        para_enthalpy = self.para.enthalpy(self.para_density, self.temperature)
        delta_h = para_enthalpy - ortho_enthalpy
        return delta_h

    # linear mixing rule for bulk hydrogen density from ortho/para
    # combo.

    def _calculate_molar_density(self):
        rho = (1-self.xp)*self.ortho_density + self.xp * \
            self.para_density        # molar density (mol/m3)
        return rho

    def _calculate_thermal_conductivity(self):
        viscosity_contributions = self.abstract_state.viscosity_contributions()
        mu_background = viscosity_contributions['residual'] - \
            viscosity_contributions['critical']

        # para k
        ph_k, _ = self.para.thermal_conductivity(
            self.para_density, self.temperature, mu_background)

        # Normal-hydrogen
        nh_k, _ = self.normal.thermal_conductivity(
            self.normal_density, self.temperature, mu_background)

        # Linear interp/extrap of k
        k = nh_k + (ph_k - nh_k) * (0.75 - self.get_xo())/0.75
        return k

    def _calculate_specific_heat_capacity(self):
        # para cp
        ph_cp = self.para.specific_heat_constant_pressure(
            self.para_density, self.temperature)

        # ortho cp
        oh_cp = self.ortho.specific_heat_constant_pressure(
            self.ortho_density, self.temperature)

        # Linear interp of cp
        # molar specific heat capacity (J/mol K)
        cp_molar = (1-self.xp)*oh_cp + self.xp*ph_cp
        cp = cp_molar / self.molecular_mass  # convert to cp mass (J/kg K)
        return cp

    def _calculate_viscosity(self):
        v = self.normal.viscosity(
            self.normal_density, self.temperature, self.molecular_mass)
        return v

    def _calculate_enthalpy(self):
        # para h
        ph_h = self.para.enthalpy(self.para_density, self.temperature)

        # ortho h
        oh_h = self.ortho.enthalpy(self.ortho_density, self.temperature)

        # Linear interp of h
        # molar specific heat capacity (J/mol)
        h_molar = (1-self.xp)*oh_h + self.xp*ph_h
        h = h_molar / self.molecular_mass  # convert to h mass (J/kg)
        return h

    def _calculate_entropy(self):
        # para s
        ph_s = self.para.entropy(self.para_density, self.temperature)

        # ortho s
        oh_s = self.ortho.entropy(self.ortho_density, self.temperature)

        # Linear interp of cp
        # molar specific heat capacity (J/mol)
        s_molar = (1-self.xp)*oh_s + self.xp*ph_s
        s = s_molar / self.molecular_mass  # convert to s mass (J/kg)
        return s

    def _calculate_critical_temperature(self):
        Tc = self.get_xo()*self.ortho.Tc + self.get_xp()*self.para.Tc
        return Tc

    def _calculate_critical_pressure(self):
        Pc = self.get_xo()*self.ortho.Pc + self.get_xp()*self.para.Pc
        return Pc


class HeliumNeon(FluidStream):
    def __init__(self, mass_flow_rate, helium_fraction):
        FluidStream.__init__(self, 'Helium&Neon', mass_flow_rate)
        self.specify_phase(CP.iphase_gas)
        self.set_mole_fractions([helium_fraction, 1 - helium_fraction])

        self.helium_as = CP.AbstractState('HEOS', 'Helium')
        self.neon_as = CP.AbstractState('HEOS', 'Neon')

        # ref temp
        self.T0 = 298.15
        self.mu_0 = np.array([19.8253, 31.7088])*1e-6
        self.lam_0 = np.array([155.0008, 49.1732])*1e-3

        # viscosity coeffs
        self.a_coeffs = np.array([[6.8257552e-1,	6.75404e-1],
                                  [1.4496203e-2,	-2.03522e-2],
                                  [1.1987706e-3,	1.61102e-2],
                                  [-6.7722412e-5,	-4.88074e-3],
                                  [4.9875650e-5,	5.32334e-4],
                                  [-6.1456994e-6,	2.93695e-4],
                                  [1.3189407e-6,	-1.55155e-4],
                                  [-3.7245774e-7,	3.10797e-5],
                                  [1.3671981e-8,	-2.50504e-6],
                                  [5.0354149e-8,	2.74563e-8],
                                  [-1.5714379e-8, 0.0],
                                  [1.4720785e-9, 0.0]])

        # thermal cond coeffs
        self.b_coeffs = np.array([[6.8192175e-1,	6.76478e-1],
                                  [1.4441872e-2,	-2.13734e-2],
                                  [1.2138429e-3,	1.63523e-2],
                                  [-7.4912205e-5,	-4.79402e-3],
                                  [5.2123986e-5,	4.23959e-4],
                                  [-6.0795764e-6,	3.35013e-4],
                                  [8.6135146e-7,	-1.53714e-4],
                                  [-2.6311453e-7,	2.45053e-5],
                                  [6.8367328e-8,	-5.03701e-7],
                                  [1.6608814e-8,	-1.67000e-7],
                                  [-9.0525341e-9, 0.0],
                                  [9.9986305e-10, 0.0]])

    def _calculate_viscosity(self):
        T = self.temperature
        inds = np.arange(start=1, stop=self.a_coeffs.shape[0]+1)
        mu_vals = np.zeros((2,))
        for f_idx in (0, 1):
            a_vals = self.a_coeffs[:, f_idx]
            terms = a_vals * np.log(T / self.T0) ** inds
            r_mu_gas = np.exp(np.sum(terms))
            mu_vals[f_idx] = r_mu_gas * self.mu_0[f_idx]

        mu = np.matmul(self.mole_fractions, mu_vals)
        return mu

    def _calculate_thermal_conductivity(self):
        T = self.temperature
        inds = np.arange(start=1, stop=self.b_coeffs.shape[0]+1)
        lam_vals = np.zeros((2,))
        for f_idx in (0, 1):
            b_vals = self.b_coeffs[:, f_idx]
            terms = b_vals * np.log(T / self.T0) ** inds
            r_lam_gas = np.exp(np.sum(terms))
            lam_vals[f_idx] = r_lam_gas * self.lam_0[f_idx]

        lam = np.matmul(self.mole_fractions, lam_vals)
        return lam

    def _calculate_critical_temperature(self):
        z = self.mole_fractions
        he_tcrit = self.helium_as.T_critical()
        ne_tcrit = self.neon_as.T_critical()
        Tc = z[0]*he_tcrit + z[1]*ne_tcrit
        return Tc

    def _calculate_critical_pressure(self):
        z = self.mole_fractions
        he_pcrit = self.helium_as.p_critical()
        ne_pcrit = self.neon_as.p_critical()
        Pc = z[0]*he_pcrit + z[1]*ne_pcrit
        return Pc



# Ergun equation
def ergun_equation(epsilon, Re_p):
    # dPdx = 150*mu/dp^2*(1-eta)^2/eta^3*vs + 1.75*rho/dp*(1-eta)/eta^3*vs*abs(vs);
    ff = 150/Re_p + 1.75
    return ff

# Hicks equation (pressure drop through packed bed)
def hicks_equation(mu, dp, epsilon, vs, Re_p):
    dPdx = mu*(1-epsilon)**2*vs/(dp**2*epsilon**3) * 6.8 * Re_p**0.8
    return dPdx


# catalyst object

class Catalyst(object):
    def __init__(self, catalyst_configuration):
        self.solid_fraction = catalyst_configuration['solid_fraction']
        self.particle_diameter = catalyst_configuration['diameter']
        self.density = catalyst_configuration['density']
        self.void_fraction = 1 - catalyst_configuration['solid_fraction']

    def calculate_catalyst_mass(self, reaction_side_volume):
        catalyst_volume = reaction_side_volume * self.solid_fraction
        catalyst_mass = catalyst_volume * self.density
        return catalyst_mass

    # correlation for T.C. of hematite (Molgaard et al "Thermal Conductivity of Magnetite and Hematite")
    # (assuming catalyst is Ionex - which is mainly Iron Oxide)
    # note this function has very minimal impact on results.
    def calculate_thermal_conductivity(self, temperature):
        k = 8.39 - 6.63e-3*temperature
        return k



# main file of hydrogen-pfhx

import yaml
import numpy as np
from scipy.integrate import solve_bvp
from hydrogen_pfhx import (fluids, catalysts, hexs, bvp_model, helium_neon, outputs, utils)

def model(configuration_file = 'src/configs/default_configuration.yaml'):
    # Step 1. read config file
    with open(configuration_file, "r") as stream:
        try:
            configuration = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    # Step 2. Create a reactant, coolant, reactor & catalyst
    # reactant
    reactant_mass_flowrate_kps = utils.tpd_to_kps(
        configuration['reactant']['mass_flow_rate'])
    reactant = fluids.Hydrogen(reactant_mass_flowrate_kps)

    # coolant
    coolant_name = configuration['coolant']['fluid']
    reactor_mass_flowrate_kps = utils.tpd_to_kps(
        configuration['coolant']['mass_flow_rate'])
    if coolant_name == 'Hydrogen':
        xp_cool_inlet = configuration['coolant']['x_para']
        coolant = fluids.Hydrogen(reactor_mass_flowrate_kps)
        coolant.update_composition(xp_cool_inlet)
    elif coolant_name == 'HeliumNeon':
        helium_neon.setup_CoolProp(['Helium', 'Neon', 'Argon'])
        helium_fraction = configuration['coolant']['helium_fraction']
        coolant = fluids.HeliumNeon(reactor_mass_flowrate_kps, helium_fraction)
    else:
        coolant = fluids.FluidStream(coolant_name, reactor_mass_flowrate_kps)

    # reactor
    reactor = hexs.PlateFinHex(configuration['reactor'])

    # catalyst
    catalyst = catalysts.Catalyst(configuration['catalyst'])

    # Step 3. Setup the boundary value problem - initialise a solution
    boundary_properties = np.array((
        configuration['reactant']['x_para'],
        configuration['reactant']['pressure'],
        configuration['reactant']['temperature'],
        configuration['coolant']['pressure'],
        configuration['coolant']['temperature']))

    (x_mesh, sol_init) = bvp_model.initialise_solution(reactant, coolant,
                                                       reactor, catalyst, boundary_properties, configuration['simulation'])

    # Step 4. Run the BVP solver
    additional_parameters = (reactor, catalyst, reactant,
                             coolant, boundary_properties)
    solution = solve_bvp(lambda x, y: bvp_model.bvp_function(x, y, additional_parameters=additional_parameters),
                         lambda xb, yb: bvp_model.counter_current_boundary_condition(
                             xb, yb, boundary_properties),
                         x_mesh, sol_init, tol=configuration['simulation']['tolerance'], max_nodes=1000, verbose=2)

    # Step 5. Post-process & plot results.
    results = outputs.post_process(solution, reactant, coolant, reactor, catalyst, boundary_properties)

    return results



import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from operator import sub
from hydrogen_pfhx import bvp_model


def post_process(solution, reactant, coolant, reactor, catalyst, boundary_conditions):
    z = solution.x
    xp_reactant = solution.y[0, :]
    P_reactant = solution.y[1, :]
    T_reactant = solution.y[2, :]
    P_coolant = solution.y[3, :]
    T_coolant = solution.y[4, :]

    rho_reactant = np.zeros(len(z),)
    rho_coolant = np.zeros(len(z),)
    xp_equil = np.zeros(len(z),)
    for ni in np.arange(len(z)):
        reactant, coolant = bvp_model.update_parameters(
            solution.y[:, ni], reactant, coolant, boundary_conditions)
        rho_reactant[ni] = reactant.mass_density
        rho_coolant[ni] = coolant.mass_density
        xp_equil[ni] = reactant.get_xp_equil()

    # calculate velocity
    u_reactant = reactant.mass_flow_rate / \
        (rho_reactant*reactor.total_hot_side_area()*catalyst.void_fraction)
    u_coolant = coolant.mass_flow_rate / \
        (rho_coolant*reactor.total_cold_side_area())

    columns = ('Z (m)', 'Reactant pressure (kPa)', 'Coolant pressure (kPa)', 'Reactant temperature (K)', 'Coolant temperature (K)',
               'Reactant velocity (m/s)', 'Coolant velocity (m/s)', 'Equilibrium para-hydrogen fraction (mol/mol)', 'Actual para-hydrogen fraction (mol/mol)')
    results_array = np.transpose(np.vstack(
        [z, P_reactant, P_coolant, T_reactant, T_coolant, u_reactant, u_coolant, xp_equil, xp_reactant]))
    results = pd.DataFrame(results_array, columns=columns)

    return results

def save_results(results, file_path='output/results.csv'):
    # save results
    if (file_path[:7] == 'output/') & (not os.path.isdir('output')):
        os.mkdir('output')
        
    results.to_csv(file_path)


def plot_results(results):
    # set figure fonts
    tfont = {'fontname': 'Times New Roman', 'fontweight': 'bold'}
    tfont_1 = {'fontname': 'Times New Roman'}
    plt.rcParams['font.family'] = 'serif'
    plt.rcParams['font.serif'] = ['Times New Roman']
    plt.rcParams['mathtext.it'] = 'Times New Roman:italic:bold'
    plt.rcParams['mathtext.cal'] = 'Times New Roman:italic'
    plt.rcParams['mathtext.default'] = 'regular'
    plt.rcParams["mathtext.fontset"] = 'custom'
    plt.rc('axes', titlesize=11)     # fontsize of the axes title
    plt.rc('axes', labelsize=11)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=9)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=9)    # fontsize of the tick labels

    labels = ('(b) Isomer conversion',
              '(a) Temperature',
              '(c) Velocity')

    ylabels = ('H$_{2,para}$ fraction (mol/mol)',
               'Temperature (K)',
               'Velocity (m/s)')

    nc = 4

    cc = [0.2, 0.3, 0.8]
    cr = [0.8, 0.3, 0.2]

    x_data = results['Z (m)']
    num_params = 3
    xlims = (0, np.max(x_data))
    l1 = 'reactant'

    nr = 1
    nc = 3
    fig, axs = plt.subplots(nr, nc, figsize=(7.2, 2.5), dpi=150)
    for p_idx in np.arange(num_params):
        if p_idx == 0:
            ci = 1
            ylims = (0, 1)
            r_data = results['Actual para-hydrogen fraction (mol/mol)']
            c_data = results['Equilibrium para-hydrogen fraction (mol/mol)']
        elif p_idx == 1:
            ci = 0
            ylims = (20, 80)
            r_data = results['Reactant temperature (K)']
            c_data = results['Coolant temperature (K)']
        elif p_idx == 2:
            ci = 2
            ylims = (0, 1.5)
            r_data = results['Reactant velocity (m/s)']
            c_data = results['Coolant velocity (m/s)']

        c1 = cr
        if p_idx == 0:
            ls = '--'
            c2 = 'k'
            l2 = 'equil.'
        else:
            ls = '-'
            c2 = cc
            l2 = 'coolant'
        curr_ax = axs[ci]
        curr_ax.plot(x_data, r_data, '-', color=c1, label=l1)
        curr_ax.plot(x_data, c_data, ls, color=c2, label=l2)

        # axis limits
        curr_ax.set_xlim(xlims)
        curr_ax.set_ylim(ylims)
        y_lim_diff = np.diff(ylims)[0]

        if p_idx == 0:
            plot_arrow(curr_ax, x_data, r_data, cr, '-|>', -1, y_lim_diff)
        elif p_idx == 1:
            plot_arrow(curr_ax, x_data, r_data, cr, '-|>', 1, y_lim_diff)
            plot_arrow(curr_ax, x_data, c_data, cc, '<|-', -1, y_lim_diff)
        elif p_idx == 2:
            plot_arrow(curr_ax, x_data, r_data, cr, '-|>', 1, y_lim_diff)
            plot_arrow(curr_ax, x_data, c_data, cc, '<|-', -1, y_lim_diff)

        # labels
        curr_ax.set_xlabel('Length along reactor (m)', **tfont)
        curr_ax.set_ylabel(ylabels[p_idx], **tfont)
        curr_ax.set_title(labels[p_idx], **tfont_1)

        # tick markers
        curr_ax.minorticks_on()
        curr_ax.tick_params(direction='in', which='minor', length=2,
                            bottom=True, top=True, left=True, right=True)
        curr_ax.tick_params(direction='in', which='major', length=4,
                            bottom=True, top=True, left=True, right=True)

        curr_ax.set_xticks(np.arange(xlims[0], xlims[1]+1, 1.0))
        if p_idx == 0:
            loc = 'lower right'
        else:
            loc = 'upper right'
        lg = curr_ax.legend(loc=loc)
        frame = lg.get_frame()
        frame.set_edgecolor('black')

    fig.tight_layout()

    fig.subplots_adjust(left=0.08, top=0.88, bottom=0.2, right=0.98)


def get_aspect(ax):
    # Total figure size
    figW, figH = ax.get_figure().get_size_inches()
    # Axis size on figure
    _, _, w, h = ax.get_position().bounds
    # Ratio of display units
    disp_ratio = (figH * h) / (figW * w)
    # Ratio of data units
    # Negative over negative because of the order of subtraction
    data_ratio = sub(*ax.get_ylim()) / sub(*ax.get_xlim())

    return disp_ratio / data_ratio


def plot_arrow(curr_ax, x_data, y_data, c, arrow, off_dir, y_lim_diff):
    cx = np.max(x_data)/2
    dx = np.max(x_data)/5
    g = np.interp(cx, x_data, np.gradient(y_data, x_data))
    cy = np.interp(cx, x_data, y_data)
    dy = dx*g

    offset = off_dir * y_lim_diff*0.06
    curr_ax.annotate('', xy=(cx+dx/2, cy+dy/2 + offset), xytext=(cx-dx/2, cy-dy/2 + offset), xycoords='data', textcoords='data',
                     arrowprops=dict(arrowstyle=arrow, color=c))



import numpy as np
import scipy.optimize
import CoolProp.CoolProp as CP

class HydrogenData(object):
    # Relevant equations for calculating hydrogen properties
    # Equations obtained from:
    # Leachman JW, Jacobsen RT, Penoncello SG, Lemmon EW.
    # Fundamental equations of state for parahydrogen, normal
    # hydrogen, and orthohydrogen. J Phys Chem Ref Data
    # 200938:721. https://doi.org/10.1063/1.3160306
    
   # properties
    # Universal gas constant
    R = 8.31446261815324 # m3.Pa / (K.mol)
    
    # EOS index values
    l = 7
    m = 9
    n = 14
    
    # value for limit
    limit_eval = 1e-12
    rho_min = 1e-12
    
    # viscosity params
    v_ai = [2.09630e-1,	-4.55274e-1,	1.43602e-1,	-3.35325e-2,	2.76981e-3]
    v_bi = [-0.187,	2.4871,	3.7151,	-11.0972,	9.0965,	-3.8292,	0.5166]
    v_ci = [6.43449673,	4.56334068e-2,	2.32797868e-1,	9.58326120e-1,	1.27941189e-1,	3.63576595e-1]
    
    # solving opts
    
    
    def __init__(self):
        # ref pressure
        self.p_ref = 101325 # Pa
        
        # initialise the reference temperature
        # t_guess = CP.PropsSI('T','P',self.p_ref,'Q',0,'Hydrogen')
        # self.t_ref = fzero(@(T) self.vapour_pressure(T) - self.p_ref, t_guess, self.opt)
        self.t_ref = 1000
                    
        # initialise the reference enthalpy
        rho_guess = CP.PropsSI('Dmolar','P',self.p_ref, 'T',self.t_ref,'Hydrogen')
        rho = self.calculate_density(self.p_ref, self.t_ref, rho_guess)
        self.h_ref = self.absolute_enthalpy(rho, self.t_ref)
                
    def pressure(self, rho, T):
        d_alphar_delta = self.delta_diff_alphar_delta(rho, T)
        p = rho*self.R*T * (1 + d_alphar_delta)
        return p
    
    def Z(self, rho, T):
        d_alphar_delta = self.delta_diff_alphar_delta(rho, T)
        z = 1 + d_alphar_delta
        return z
    
    # Correlation from Muzny et al.
    def viscosity(self, rho, T, M):
        # Dilute gas viscosity contribution
        omega = 0.297 # nm
        e_kb = 30.41 # K
        T_star = T/e_kb
        S_star = np.exp(np.sum(self.v_ai*np.log(T_star)**np.arange(start=0,stop=len(self.v_ai))))
        mu_0 = 0.021357*(M*1e3*T)**0.5/(omega**2.0*S_star)*1e-6
        
        # initial density contribution
        B_mu_star = np.sum(self.v_bi*T_star**(-1*np.arange(start=0,stop=len(self.v_bi))))
        B_mu = B_mu_star * omega **3
        mu_1 = B_mu * mu_0
        mu_init = mu_1 * rho * M
        
        # additional contribution (incorporates both mu_h & mu_c, which
        # are not segregated in Muzny et al.)
        # only use normal hydrogen Tc as that is what is used in paper.
        rho_sc = 90.5 # kg/m3
        Tc = 33.145 # K
        rho_r = rho * M / rho_sc
        T_r = T / Tc
        
        mu_add = 1e-6 * self.v_ci[0] * rho_r**2.0 * np.exp(self.v_ci[1]*T_r + self.v_ci[2]/T_r + self.v_ci[3]*rho_r**2/(self.v_ci[4] + T_r) + self.v_ci[5]*rho_r**6)
        
        mu = mu_0 + mu_init + mu_add
        return mu
    
    def enthalpy(self, rho, T):
        h = self.absolute_enthalpy(rho, T) - self.h_ref
        return h
    
    def entropy(self, rho, T):
        s = self.R * (self.tau_diff_alpha0_tau(T) + self.tau_diff_alphar_tau(rho, T) - self.ideal_helmholtz_energy(rho, T) - self.residual_helmholtz_energy(rho, T))
        return s
    
    # Enthalpy will be in J/mol
    def absolute_enthalpy(self, rho, T):
        h = self.R*T * (1 + self.tau_diff_alpha0_tau(T) + self.tau_diff_alphar_tau(rho, T) + self.delta_diff_alphar_delta(rho, T))
        return h
    
    def specific_heat_constant_volume(self, rho, T):
        tau2_d2_alpha0_dtau2 = self.tau2_diff2_alpha0_difftau2(T)
        tau2_d2_alphar_dtau2 = self.tau2_diff2_alphar_difftau2(rho, T)
        cv = -self.R * (tau2_d2_alpha0_dtau2 + tau2_d2_alphar_dtau2)
        return cv
    
    def specific_heat_constant_pressure(self, rho, T):
        cv = self.specific_heat_constant_volume(rho, T)
        numerator = 1 + self.delta_diff_alphar_delta(rho, T) - self.delta_tau_diff2_alphar_diff_delta_tau(rho,T)
        denominator = 1 + 2.0*self.delta_diff_alphar_delta(rho, T) + self.delta2_diff2_alphar_diffdelta2(rho,T)
        cp = cv + self.R*numerator**2/denominator
        return cp
    
    def speed_of_sound(self, rho, T):
        u = self.R*T * (self.tau_diff_alpha0_tau(T) + self.tau_diff_alphar_tau(rho, T))
        return u

    def pressure_error(self, rho, T, P):
        pe = self.pressure(rho, T) - P
        return pe
    
    def calculate_density(self, P, T, rho_guess):
        rho = scipy.optimize.fsolve(self.pressure_error,  x0 = rho_guess, args=(T, P), xtol=1e-6)
        return rho
         
    # Correlation & data from Assael et al.
    def thermal_conductivity(self, rho, T, mu_background):
        if type(self) == 'modelling.hydrogen_thermodynamics.OrthoHydrogen':
            print('thermal conductivity data is not available for ortho-hydrogen - define as mixture of para/normal')
        else:
            # reduced temperature & density
            rho_r = rho / self.rho_c
            T_r = T / self.Tc
            
            # dilute gas thermal conductivity
            ai_m = np.arange(len(self.tc_A1))
            ai_n = np.arange(len(self.tc_A2))
            lambda_0 = np.sum(self.tc_A1 * T_r**ai_m) / np.sum(self.tc_A2 * T_r**ai_n)
            
            # excess thermal conductivity
            bi = np.arange(len(self.tc_B1))
            lambda_excess = np.sum((self.tc_B1 + self.tc_B2 * T_r) * rho_r**bi)
            
            # empirical critical enhancement
            model = 'olchowy_sengers'
            if model == 'none':
                lambda_c = 0
            elif model == 'empirical':
                delta_Tc = T_r - 1
                delta_rho_r = rho_r - 1
                lambda_c = self.tc_C[0] / (self.tc_C[1] + abs(delta_Tc)) * np.exp(-(self.tc_C[2]*delta_rho_r)**2)
            elif model == 'olchowy_sengers':
                # Olchowy and Sengers model
                cp = self.specific_heat_constant_pressure(rho, T)
                cv = self.specific_heat_constant_volume(rho, T)
                
                # constants
                kb = 1.38064852e-23 # m2 kg s-2 K-1
                RD = 1.01
                nu = 0.63
                gamma = 1.2415
                GAMMA = 0.172
                xi_0 = 1.5e-10
                T_ref = 1.5 * self.Tc
                
                # Eq 7, 6 & 5 (Assael et al)
                del_psi = self.R *self.Tc * rho / self.rho_c * (self.diff_rho_pressure(rho, T) - T_ref/T * self.diff_rho_pressure(rho, T_ref))
                min_val = 1e-16
                xi = xi_0 * (GAMMA**(-1) * np.max([min_val,del_psi])) ** (nu/gamma)
                y = self.qd*xi
                omega_0 = 2/np.pi * (1 - np.exp(-1/((y)**(-1.0) + (y*self.rho_c/rho)**2/3)))
                omega = 2/np.pi * ((cp - cv)/cp * np.arctan(y) + cv/cp*y)
                
                if xi > 0:
                    lambda_c = rho * cp * RD * kb * T /(6*np.pi*mu_background*xi) * (omega - omega_0)
                else:
                    lambda_c = 0
                
                lambda_c = np.max([0, lambda_c])
            
            
            lambda_val = lambda_0 + lambda_excess + lambda_c
    
        return lambda_val, lambda_c
    
    # def for specific heat capacity (Eq 25)
    def ideal_specific_heat_capacity(self, T):
        cp = self.R * (2.5 + np.sum(self.uk * (self.vk/T)**2.0 * np.exp(self.vk/T) / (np.exp(self.vk/T) - 1)**2))
        return cp
    
    # def for ideal component of Helmholtz energy (Eq 31)
    def ideal_helmholtz_energy(self, rho, T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        alpha_0 = np.log(delta) + 1.5 * np.log(tau) + self.ak[0] + self.ak[1]*tau +\
            np.sum(self.ak[2:] * np.log(1 - np.exp(self.bk[2:]*tau)))
        return alpha_0
    
    # def for residual component of Helmholtz energy (Eq 32)
    def residual_helmholtz_energy(self, rho, T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        main_terms = self.Ni_eos * (delta ** self.di) * (tau ** self.ti)
        gaussian_terms = self.phi_i *(delta - self.D_i)**2.0 + self.beta_i * (tau - self.gamma_i)**2
        alpha_r = np.sum(main_terms[:self.l]) + \
            np.sum(main_terms[self.l:self.m]*np.exp(-(delta) ** self.pi[self.l:self.m])) + \
            np.sum(main_terms[self.m:self.n]*np.exp(-gaussian_terms))
        return alpha_r
    
    # combined helmholtz energy (Eq 20)
    def reduced_helmholtz_energy(self, rho, T):
        alpha = self.ideal_helmholtz_energy(rho, T) + self.residual_helmholtz_energy(rho, T)
        return alpha
    
    # final Helmholtz energy (not reduced) (Eq 17)
    def helmholtz_energy(self, rho, T):
        a = self.reduced_helmholtz_energy(rho, T) * self.R * T
        return a
    
    def vapour_pressure(self, T):
        tau = self.Tc / T
        theta = 1 - T / self.Tc
        p_omega = self.Pc * np.exp(tau * np.sum(self.Ni_vp * theta ** self.ki))
        return p_omega
    
    def second_virial_coefficient(self, T):
        # syms rho
        f = 1 / self.rho_c*self.delta_diff_alphar_delta(self.rho_min, T) # @(rho) 
        B = f #(self.limit_eval) #eval(limit(f, rho, 0,'right'))
        return B
    
    def third_virial_coefficient(self, T):
        # syms rho
        f = 1 / self.rho_c()**2.0 * self.delta2_diff2_alphar_diffdelta2(self.rho_min, T) # @(rho)
        C = f #(self.limit_eval) #eval(limit(f, rho, 0,'right'))
        return C
    
    def diff_rho_pressure(self, rho, T):
        # at constant T
        drho_dp = 1/self.diff_pressure_rho(rho, T)
        return drho_dp
    
    def diff_pressure_rho(self, rho, T):
        # (at constant T)
        dp_drho = self.R * T * (1 + 2.0*self.delta_diff_alphar_delta(rho, T) + self.delta2_diff2_alphar_diffdelta2(rho, T))
        return dp_drho
    
    def delta_diff_alphar_delta(self, rho, T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        main_terms = self.Ni_eos * delta ** self.di * tau ** self.ti
        extra_terms = (self.di - self.pi * delta**self.pi)
        gaussian_terms = self.phi_i *(delta - self.D_i)**2.0 + self.beta_i * (tau - self.gamma_i)**2
        extra_gaussian_terms = self.di[self.m:self.n] - 2.0*delta*self.phi_i*(delta - self.D_i)
        d_alphar_delta = np.sum(self.di[:self.l]*main_terms[:self.l]) + \
            np.sum(main_terms[self.l:self.m]*np.exp(-(delta) ** self.pi[self.l:self.m])*extra_terms[self.l:self.m]) + \
            np.sum(main_terms[self.m:self.n]*np.exp(-gaussian_terms)*extra_gaussian_terms)
        return d_alphar_delta
    
    def tau_diff_alpha0_tau(self, T):
        tau = self.Tc / T
        d_alpha0_tau = 1.5 + tau*self.ak[1] +\
            tau*np.sum(self.ak[2:] * -1.0*self.bk[2:]*np.exp(self.bk[2:]*tau) / (1 - np.exp(self.bk[2:]*tau)))
        return d_alpha0_tau
    
    def tau_diff_alphar_tau(self, rho, T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        main_terms = self.Ni_eos * delta ** self.di * tau ** self.ti
        extra_terms = self.ti[self.l:self.m]
        gaussian_terms = self.phi_i *(delta - self.D_i)**2.0 + self.beta_i * (tau - self.gamma_i)**2
        extra_gaussian_terms = self.ti[self.m:self.n] - 2.0*tau*self.beta_i*(tau - self.gamma_i)
        d_alphar_tau = np.sum(self.ti[:self.l]*main_terms[:self.l]) + \
            np.sum(main_terms[self.l:self.m]*np.exp(-(delta) ** self.pi[self.l:self.m])*extra_terms) + \
            np.sum(main_terms[self.m:self.n]*np.exp(-gaussian_terms)*extra_gaussian_terms)
        return d_alphar_tau
    
    def tau2_diff2_alpha0_difftau2(self, T):
        tau = self.Tc / T
        eff_bk = np.array(self.bk[2:])
        d2_alpha0_dtau2 = -1.5 - tau**2.0 *np.sum(self.ak[2:] * eff_bk**2.0 * np.exp(eff_bk*tau) * ((np.exp(eff_bk*tau) - 1.0)**(-2.0)))
        return d2_alpha0_dtau2
    
    def tau2_diff2_alphar_difftau2(self, rho, T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        main_terms = self.Ni_eos * delta ** self.di * tau ** self.ti
        ti_array = np.array(self.ti)
        extra_terms = ti_array * (np.array(self.ti) - 1.0)
        gaussian_terms = self.phi_i *(delta - self.D_i)**2.0 + self.beta_i * (tau - self.gamma_i)**2
        extra_gaussian_terms = (ti_array[self.m:self.n] - 2.0 * np.array(self.beta_i) * tau*(tau - np.array(self.gamma_i)))**2.0 - \
            ti_array[self.m:self.n] - 2.0*tau**2.0 *np.array(self.beta_i)
        d2_alphar_dtau2 = np.sum(extra_terms[:self.l]*main_terms[:self.l]) + \
            np.sum(main_terms[self.l:self.m]*np.exp(-(delta) ** self.pi[self.l:self.m])*extra_terms[self.l:self.m]) + \
            np.sum(main_terms[self.m:self.n]*np.exp(-gaussian_terms)*extra_gaussian_terms)
        return d2_alphar_dtau2
    
    def delta_tau_diff2_alphar_diff_delta_tau(self,rho,T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        main_terms = self.Ni_eos * delta ** self.di * tau ** self.ti
        li = self.pi[self.l:self.m]
        extra_terms = (self.di[self.l:self.m] - li * (delta) ** li) * self.ti[self.l:self.m]
        gaussian_terms = self.phi_i *(delta - self.D_i)**2.0 + self.beta_i * (tau - self.gamma_i)**2.0
        extra_gaussian_terms = (self.di[self.m:self.n] - 2.0 * np.array(self.phi_i) * delta *(delta - np.array(self.D_i))) * \
            (self.ti[self.m:self.n] - 2.0 * np.array(self.beta_i) * tau*(tau - np.array(self.gamma_i)))
        d2_alphar_delta_tau = np.sum(np.array(self.di[:self.l]) * np.array(self.ti[:self.l]) * main_terms[:self.l]) + \
            np.sum(main_terms[self.l:self.m]*np.exp(-(delta) ** li)*extra_terms) + \
            np.sum(main_terms[self.m:self.n]*np.exp(-gaussian_terms)*extra_gaussian_terms)
        return d2_alphar_delta_tau
    
    def delta2_diff2_alphar_diffdelta2(self,rho,T):
        delta = rho / self.rho_c
        tau = self.Tc / T
        main_terms = self.Ni_eos * delta ** self.di * tau ** self.ti
        li = self.pi[self.l:self.m]
        di_array = np.array(self.di)
        extra_terms = (di_array[self.l:self.m] - li * delta ** li) * \
            (di_array[self.l:self.m] - 1.0 - li * delta ** li) - \
            li**2.0 * delta ** li
        gaussian_terms = self.phi_i *(delta - self.D_i)**2.0 + self.beta_i * (tau - self.gamma_i)**2.0
        extra_gaussian_terms = (self.di[self.m:self.n] - 2.0 * self.phi_i * delta *(delta - self.D_i))**2.0 - \
            self.di[self.m:self.n] - 2.0*delta**2.0 *self.phi_i
        d2_alphar_ddelta2 = np.sum(self.di[:self.l] * (self.di[:self.l] - 1) * main_terms[:self.l]) + \
            np.sum(main_terms[self.l:self.m]*np.exp(-(delta) ** li)*extra_terms) + \
            np.sum(main_terms[self.m:self.n]*np.exp(-gaussian_terms)*extra_gaussian_terms)
            
        return d2_alphar_ddelta2
    
    
class NormalHydrogen(HydrogenData):
    # property data for Normal hydrogen (75# para, 25# ortho)
    # Data obtained from:
    # Leachman JW, Jacobsen RT, Penoncello SG, Lemmon EW.
    # Fundamental equations of state for parahydrogen, normal
    # hydrogen, and orthohydrogen. J Phys Chem Ref Data
    # 200938:721. https://doi.org/10.1063/1.3160306
    def __init__(self):
        # specific heat capacity coefficient data
        self.uk = [1.616, -0.4117, -0.792, 0.758, 1.217]
        self.vk = [531, 751, 1989, 2484, 6859]

        # ideal gas heat capacity coefficients
        self.ak = np.array([-1.457985648,	1.888076782,	1.616,	-0.4117,	-0.792,	0.758,	1.217])
        self.bk = np.array([np.nan,	np.nan,	-16.02051591,	-22.6580178,	-60.00905114,	-74.94343038,	-206.9392065])
        
        # Polynomial terms for hydrogen EOS
        self.Ni_eos = [-6.93643,	0.01,	2.11101,	4.52059,	0.732564,	-1.34086,	0.130985,	-0.777414,	0.351944,	-0.0211716,	0.0226312,	0.032187,	-0.0231752,	0.0557346]
        self.ti = np.array([0.6844,	1.0,	0.989,	0.489,	0.803,	1.1444,	1.409,	1.754,	1.311,	4.187,	5.646,	0.791,	7.249,	2.986])
        self.di = np.array([1,	4,	1,	1,	2,	2,	3,	1,	3,	2,	1,	3,	1,	1])
        self.pi = np.array([0,	0,	0,	0,	0,	0,	0,	1,	1,	np.nan, np.nan, np.nan, np.nan, np.nan])
        
        # Gaussian terms for hydrogen EOS
        self.phi_i = np.array([1.685,	0.489,	0.103,	2.506,	1.607])
        self.beta_i = np.array([0.171,	0.2245,	0.1304,	0.2785,	0.3967])
        self.gamma_i = np.array([0.7164,	1.3444,	1.4517,	0.7204,	1.5445])
        self.D_i = np.array([1.506,	0.156,	1.736,	0.67,	1.662])
        
        # Critical properties
        self.Tc = 33.145 # K
        self.Pc = 1.2964e6 # Pa
        self.rho_c = 15.508e3 # mol/m3
        
        # Vapour pressure properties
        self.Ni_vp = [-4.89789,	0.988558,	0.349689,	0.499356]
        self.ki = [1, 1.5, 2, 2.85]
        
        # Thermal conductivity parameters (from Assael et al.)
        self.tc_A1 = np.array([-3.40976E-01,	4.58820E+00,	-1.45080E+00,	3.26394E-01,	3.16939E-03,	1.90592E-04,	-1.13900E-06])
        self.tc_A2 = np.array([1.38497E+02,	-2.21878E+01,	4.57151E+00,	1.00000E+00])
							
        self.tc_B1 = np.array([3.63081E-02,	-2.07629E-02,	3.14810E-02,	-1.43097E-02,	1.74980E-03])
        self.tc_B2 = np.array([1.83370E-03,	-8.86716E-03,	1.58260E-02,	-1.06283E-02,	2.80673E-03])
							
        self.tc_C = [6.24e-4, -2.58e-7, 0.837]
        
        self.qd = (4e-10)**(-1.0)
        
        # call superclass init
        HydrogenData.__init__(self)


class OrthoHydrogen(HydrogenData):
    # property data for Orthohydrogen
    # Data obtained from:
    # Leachman JW, Jacobsen RT, Penoncello SG, Lemmon EW.
    # Fundamental equations of state for parahydrogen, normal
    # hydrogen, and orthohydrogen. J Phys Chem Ref Data
    # 200938:721. https://doi.org/10.1063/1.3160306
    def __init__(self):
        # specific heat capacity coefficient data
        self.uk = [2.54151, -2.3661, 1.00365, 1.22447]
        self.vk = [856, 1444, 2194, 6968]

        # ideal gas heat capacity coefficients
        self.ak = np.array([-1.4675442336, 1.8845068862, 2.54151, -2.3661, 1.00365, 1.22447])
        self.bk = np.array([np.nan, np.nan, -25.7676098736, -43.4677904877, -66.0445514750, -209.7531607465])
        
        # Polynomial terms for hydrogen EOS
        self.Ni_eos = [-6.83148,		0.01,		2.11505,		4.38353,		0.211292,		-1.00939,		0.142086,	-0.87696,	0.804927,	-0.710775,	0.0639688,	0.0710858,	-0.087654,	0.647088]
        self.ti = np.array([0.7333,	1.0,	1.1372,	0.5136,	0.5638,	1.6248,	1.829,	2.404,	2.105,	4.1,	7.658,	1.259,	7.589,	3.946])
        self.di = np.array([1,	4,	1,	1,	2,	2,	3,	1,	3,	2,	1,	3,	1,	1])
        self.pi = np.array([0,	0,	0,	0,	0,	0,	0,	1,	1,	np.nan, np.nan, np.nan, np.nan, np.nan])
        
        # Gaussian terms for hydrogen EOS
        self.phi_i = np.array([1.169,	0.894,	0.04,	2.072,	1.306])
        self.beta_i = np.array([0.4555,	0.4046,	0.0869,	0.4415,	0.5743])
        self.gamma_i = np.array([1.5444,	0.6627,	0.763,	0.6587,	1.4327])
        self.D_i = np.array([0.6366,	0.3876,	0.9437,	0.3976,	0.9626])
        
        # Critical properties
        self.Tc = 33.22 # K
        self.Pc = 1.31065*1e6 # Pa
        self.rho_c = 15.445e3 # mol/m3
        
        # Vapour pressure properties
        self.Ni_vp = [-4.88684,	1.0531,	0.856947,	-0.185355]
        self.ki = [1,	1.5,	2.7,	6.2]
        
        # call superclass init
        HydrogenData.__init__(self)


class ParaHydrogen(HydrogenData):
    # property data for Parahydrogen
    # Data obtained from:
    # Leachman JW, Jacobsen RT, Penoncello SG, Lemmon EW.
    # Fundamental equations of state for parahydrogen, normal
    # hydrogen, and orthohydrogen. J Phys Chem Ref Data
    # 200938:721. https://doi.org/10.1063/1.3160306
    def __init__(self):
        # specific heat capacity coefficient data
        self.uk = [4.30256, 	13.0289, 	-47.7365, 	50.0013, 	-18.6261, 	0.993973, 	0.536078]
        self.vk = [499, 	826.5, 	970.8, 	1166.2, 	1341.4, 	5395, 	10185]

        # ideal gas heat capacity coefficients
        self.ak = np.array([-1.4485891134, 1.884521239, 	4.30256, 	13.0289, 	-47.7365, 	50.0013, 	-18.6261, 	0.993973, 	0.536078])
        self.bk = np.array([np.nan, np.nan, -15.14967515, 	-25.09259821, 	-29.47355638, 	-35.40591414, 	-40.72499848, 	-163.79258, 	-309.2173174])
        
        # Polynomial terms for hydrogen EOS
        self.Ni_eos = [-7.33375, 	0.01, 	2.60375, 	4.66279, 	0.68239, 	-1.47078, 	0.135801, 	-1.05327, 	0.328239, 	-0.0577833, 	0.0449743, 	0.0703464, 	-0.0401766, 	0.11951]
        self.ti = np.array([0.6855, 	1.0, 	1.0, 	0.489, 	0.774, 	1.133, 	1.386, 	1.619, 	1.162, 	3.96, 	5.276, 	0.99, 	6.791, 	3.19])
        self.di = np.array([1, 	4, 	1, 	1, 	2, 	2, 	3, 	1, 	3, 	2, 	1, 	3, 	1, 	1])
        self.pi = np.array([0, 	0, 	0, 	0, 	0, 	0, 	0, 	1, 	1, np.nan, np.nan, np.nan, np.nan, np.nan])
        
        # Gaussian terms for hydrogen EOS
        self.phi_i = np.array([1.7437, 	0.5516, 	0.0634, 	2.1341, 	1.777])
        self.beta_i = np.array([0.194, 	0.2019, 	0.0301, 	0.2383, 	0.3253])
        self.gamma_i = np.array([0.8048, 	1.5248, 	0.6648, 	0.6832, 	1.493])
        self.D_i = np.array([1.5487, 	0.1785, 	1.28, 	0.6319, 	1.7104])
        
        # Critical properties
        self.Tc = 32.938 # K
        self.Pc = 1.2858*1e6 # Pa
        self.rho_c = 15.538e3 # mol/m3
        
        # Vapour pressure properties
        self.Ni_vp = [-4.87767, 	1.03359, 	0.82668, 	-0.129412]
        self.ki = [1, 	1.5, 	2.65, 	7.4]
        
        # Thermal conductivity parameters (from Assael et al.)
        self.tc_A1 = np.array([-1.24500E+00, 	3.10212E+02, 	-3.31004E+02, 	2.46016E+02, 	-6.57810E+01, 	1.08260E+01, 	-5.19659E-01, 	1.43979E-02])
        self.tc_A2 = np.array([1.42304E+04, 	-1.93922E+04, 	1.58379E+04, 	-4.81812E+03, 	7.28639E+02, 	-3.57365E+01, 	1.00000E+00])

        self.tc_B1 = np.array([2.65975E-02, 	-1.33826E-03, 	1.30219E-02, 	-5.67678E-03, 	-9.23380E-05])
        self.tc_B2 = np.array([-1.21727E-03, 	3.66663E-03, 	3.88715E-03, 	-9.21055E-03, 	4.00723E-03])

        self.tc_C = [3.57e-4, -2.46e-2, 	0.2]
        
        self.qd = (5e-10)**(-1.0)
        
        # call superclass init
        HydrogenData.__init__(self)

    